@@ -1,4 +1,3 @@
-/* -*- Mode: C; c-basic-offset: 4; indent-tabs-mode: nil -*- */
 /*
    Copyright (C) 2010 Red Hat, Inc.
 
@@ -994,6 +993,7 @@ static int create_canvas(SpiceChannel *channel, display_surface *surface)
         if (c->primary) {
             if (c->primary->width == surface->width &&
                 c->primary->height == surface->height) {
+                g_free(surface);
                 CHANNEL_DEBUG(channel, "Reusing existing primary surface");
                 return 0;
             }
@@ -1971,7 +1971,7 @@ static void display_handle_surface_create(SpiceChannel *channel, SpiceMsgIn *in)
         create_canvas(channel, surface);
         if (c->mark_false_event_id != 0) {
             g_source_remove(c->mark_false_event_id);
-            c->mark_false_event_id = FALSE;
+            c->mark_false_event_id = 0;
         }
     } else {
         surface->primary = false;
