@@ -1,4 +1,3 @@
-/* -*- Mode: C; c-basic-offset: 4; indent-tabs-mode: nil -*- */
 /*
    Copyright (C) 2010 Red Hat, Inc.
 
@@ -197,6 +196,7 @@ enum {
     PROP_SECURE_CHANNELS,
     PROP_SHARED_DIR,
     PROP_SHARE_DIR_RO,
+    PROP_WEBDAV_SERVER,
     PROP_USERNAME,
     PROP_UNIX_PATH,
     PROP_PREF_COMPRESSION,
@@ -706,6 +706,9 @@ static void spice_session_get_property(GObject    *gobject,
     case PROP_SHARE_DIR_RO:
         g_value_set_boolean(value, s->share_dir_ro);
         break;
+    case PROP_WEBDAV_SERVER:
+        g_value_set_object(value, spice_session_get_webdav_server(session));
+        break;
     case PROP_PREF_COMPRESSION:
         g_value_set_enum(value, s->preferred_compression);
         break;
@@ -1485,6 +1488,27 @@ static void spice_session_class_init(SpiceSessionClass *klass)
                               G_PARAM_CONSTRUCT |
                               G_PARAM_STATIC_STRINGS));
 
+    /**
+     * SpiceSession:webdav-server:
+     *
+     * Phodav server that is internally used by #SpiceSession to make
+     * #SpiceSession:shared-dir available to the host.
+     *
+     * Since: 0.39
+     **/
+    g_object_class_install_property
+        (gobject_class, PROP_WEBDAV_SERVER,
+         g_param_spec_object("webdav-server",
+                             "WebDAV server",
+                             "PhodavServer object used for directory sharing",
+#ifdef USE_PHODAV
+                             PHODAV_TYPE_SERVER,
+#else
+                             G_TYPE_OBJECT,
+#endif
+                             G_PARAM_READABLE |
+                             G_PARAM_STATIC_STRINGS));
+
     /**
      * SpiceSession:preferred-compression:
      *
@@ -2667,17 +2691,30 @@ static void spice_session_set_shared_dir(SpiceSession *session, const gchar *dir
 
     g_free(s->shared_dir);
     s->shared_dir = g_strdup(dir);
+
+#ifdef HAVE_PHODAV_VIRTUAL
+    if (s->webdav == NULL) {
+        return;
+    }
+
+    PhodavVirtualDir *root;
+    g_object_get(s->webdav, "root-file", &root, NULL);
+    phodav_virtual_dir_root_set_real(root, s->shared_dir);
+    g_object_unref(root);
+#endif
 }
 
 G_GNUC_INTERNAL
 const gchar* spice_audio_data_mode_to_string(gint mode)
 {
+    G_GNUC_BEGIN_IGNORE_DEPRECATIONS
     static const char *str[] = {
         [ SPICE_AUDIO_DATA_MODE_INVALID ] = "invalid",
         [ SPICE_AUDIO_DATA_MODE_RAW ] = "raw",
         [ SPICE_AUDIO_DATA_MODE_CELT_0_5_1 ] = "celt",
         [ SPICE_AUDIO_DATA_MODE_OPUS ] = "opus",
     };
+    G_GNUC_END_IGNORE_DEPRECATIONS
     return (mode >= 0 && mode < G_N_ELEMENTS(str)) ? str[mode] : "unknown audio codec";
 }
 
@@ -2807,21 +2844,39 @@ PhodavServer* spice_session_get_webdav_server(SpiceSession *session)
     static GMutex mutex;
 
     const gchar *shared_dir = spice_session_get_shared_dir(session);
+    /* with HAVE_PHODAV_VIRTUAL, PhodavServer must be created even if shared_dir is NULL */
+#ifndef HAVE_PHODAV_VIRTUAL
     if (shared_dir == NULL) {
         SPICE_DEBUG("No shared dir set, not creating webdav server");
         return NULL;
     }
+#endif
 
     g_mutex_lock(&mutex);
 
     if (priv->webdav == NULL) {
+#ifdef HAVE_PHODAV_VIRTUAL
+        PhodavVirtualDir *root = phodav_virtual_dir_new_root();
+        priv->webdav = phodav_server_new_for_root_file(G_FILE(root));
+
+        phodav_virtual_dir_root_set_real(root, shared_dir);
+
+        g_object_unref(phodav_virtual_dir_new_dir(root, SPICE_WEBDAV_CLIPBOARD_FOLDER_PATH, NULL));
+        g_object_unref(root);
+#else
         priv->webdav = phodav_server_new(shared_dir);
+#endif
+
         g_object_bind_property(session,  "share-dir-ro",
                                priv->webdav, "read-only",
                                G_BINDING_SYNC_CREATE|G_BINDING_BIDIRECTIONAL);
+
+        /* with HAVE_PHODAV_VIRTUAL, the update is done in spice_session_set_shared_dir() */
+#ifndef HAVE_PHODAV_VIRTUAL
         g_object_bind_property(session,  "shared-dir",
                                priv->webdav, "root",
                                G_BINDING_SYNC_CREATE|G_BINDING_BIDIRECTIONAL);
+#endif
     }
 
     g_mutex_unlock(&mutex);
