@@ -56,6 +56,9 @@ typedef uint8_t BYTE;
 #define MINwminext 1
 #define MAXwminext 100000000
 
+/* Maximum image size in pixels, mainly to avoid possible integer overflows */
+#define SPICE_MAX_IMAGE_SIZE (512 * 1024 * 1024 - 1)
+
 typedef struct QuicFamily {
     unsigned int nGRcodewords[MAXNUMCODES];      /* indexed by code number, contains number of
                                                     unmodified GR codewords in the code */
@@ -334,7 +337,7 @@ static void correlate_init(QuicFamily *family, int bpc)
     }
 }
 
-static void golomb_coding_slow(QuicFamily *family, const BYTE n, const unsigned int l,
+static void golomb_coding_slow(const QuicFamily *family, const BYTE n, const unsigned int l,
                                unsigned int * const codeword,
                                unsigned int * const codewordlen)
 {
@@ -564,7 +567,7 @@ static int decode_state_run(Encoder *encoder, CommonState *state)
         decode_eatbits(encoder, state->melclen);
     }
 
-    /* adjust melcoder parameters */
+    /* adjust melcorder parameters */
     if (state->melcstate) {
         state->melclen = J[--state->melcstate];
         state->melcorder = (1U << state->melclen);
@@ -1136,7 +1139,7 @@ int quic_decode_begin(QuicContext *quic, uint32_t *io_ptr, unsigned int num_io_w
     int channels;
     int bpc;
 
-    if (!encoder_reset(encoder, io_ptr, io_ptr_end)) {
+    if (!num_io_words || !encoder_reset(encoder, io_ptr, io_ptr_end)) {
         return QUIC_ERROR;
     }
 
@@ -1165,6 +1168,16 @@ int quic_decode_begin(QuicContext *quic, uint32_t *io_ptr, unsigned int num_io_w
     height = encoder->io_word;
     decode_eat32bits(encoder);
 
+    if (width <= 0 || height <= 0) {
+        encoder->usr->warn(encoder->usr, "invalid size\n");
+        return QUIC_ERROR;
+    }
+
+    /* avoid too big images */
+    if ((uint64_t) width * height > SPICE_MAX_IMAGE_SIZE) {
+        encoder->usr->error(encoder->usr, "image too large\n");
+    }
+
     quic_image_params(encoder, type, &channels, &bpc);
 
     if (!encoder_reset_channels(encoder, channels, width, bpc)) {
